# -*- coding: utf-8 -*-
"""
Django Tagulous - Fabulous Tags
"""
__version__ = "1.2.1"
__license__ = "BSD"
__author__ = "Richard Terry"
__credits__ = [
    "Dmitry Ivanchenko (ivanchenkodmitry)",
    "Erik Van Kelst (4levels)",
    "Frank Lanitz (frlan)",
    "Gaël Utard (gutard)",
    "James Pic (jpic)",
    "Khoa Pham (phamk)",
    "Mark London",
    "Martín R. Guerrero (slackmart)",
    "Mary (minidietcoke)",
    "Michael Röttger (mcrot)",
    "Pamela McA'Nulty (PamelaM)",
    "Peter Baumgartner (ipmb)",
    "Raniere Silva (rgaics)",
    "Robert Erb (rerb) at AASHE",
    "hagsteel",
    "Diego Ubirajara (dubirajara)",
    "Andrew O'Brien (marxide)",
]

default_app_config = "tagulous.apps.TagulousConfig"
